#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "defns.h"
#include "comparison.h"
#include "sortingAlgorithms.h"

#define MAX_DATA_SIZE 2000000		// Array size for the n.log(n) sort algorithms
#define MAX_DATA_SIZE_Q 50000		// Array size for the quadratic algorithms

///////////////////////////////////////////////////////////////////
// Functions Forward Declaration
///////////////////////////////////////////////////////////////////
void runSortsUsingRandomValues(void);
void runSortsUsingSortedValues(void);
void printArray(Type sortedData[], int size);
void checkSorting(void);

///////////////////////////////////////////////////////////////////
// Main Function
///////////////////////////////////////////////////////////////////
int main(void)
{

	checkSorting();
	runSortsUsingRandomValues();
	runSortsUsingSortedValues();

	return 0;
}

///////////////////////////////////////////////////////////////////
// Varius Test Functions
///////////////////////////////////////////////////////////////////

// Print the contents of an array
void printArray(Type sortedData[], int size)
{
	int i = 0;
	for (i = 0; i < size; i++) {
		printf(TYPE_MODIFIER" ", sortedData[i]);
	}
	printf("\n");
}

// Sort some data using all the algorithms
void checkSorting(void)
{
	Type initialData[40];
	Type sortedData[40];

	BOOL (*compareFunc)(Type a, Type b) = intLessThan;

	int i = 0;

	for (i = 0; i < 40; i++) {
        initialData[i] = rand() % 100;
	}

	// insertion sort
	memcpy((void *)sortedData, (const void *)initialData, 40 * sizeof(Type));
	insertionSort(sortedData, 40, compareFunc);
	printArray(sortedData, 40);

	// selection sort
	memcpy((void *)sortedData, (const void *)initialData, 40 * sizeof(Type));
	selectionSort(sortedData, 40, compareFunc);
	printArray(sortedData, 40);

	// quick sort
	memcpy((void *)sortedData, (const void *)initialData, 40 * sizeof(Type));
	quickSort(sortedData, 40, compareFunc);
	printArray(sortedData, 40);

	// heap sort
	memcpy((void *)sortedData, (const void *)initialData, 40 * sizeof(Type));
	heapSort(sortedData, 40, compareFunc);
	printArray(sortedData, 40);

	// merge sort
	memcpy((void *)sortedData, (const void *)initialData, 40 * sizeof(Type));
	mergeSort(sortedData, 40, compareFunc);
	printArray(sortedData, 40);
}


// Run the algorithms on various arrays with random values and time them
void runSortsUsingRandomValues(void)
{
	int i = 0;
	Type value = 0;
	Type *initialData = NULL;
	Type *sortedData = NULL;
	clock_t start = 0;
	clock_t duration = 0;
	int size = 0;
	BOOL (*compareFunc)(Type a, Type b) = intGreaterThan;

	// Allocate space for the arrays
	initialData = (Type *)calloc(MAX_DATA_SIZE, sizeof(Type));
	if (initialData == NULL) {
		fprintf(stderr, "Memory Allocation Error!!!\n");
		exit(1);
	}
	sortedData = (Type *)calloc(MAX_DATA_SIZE, sizeof(Type));
	if (sortedData== NULL) {
		fprintf(stderr, "Memory Allocation Error!!!\n");
		exit(1);
	}

	// Initialize the array with random values
	srand((unsigned int)time(NULL));
	for (i = 0; i < MAX_DATA_SIZE; i++) {
        initialData[i] = rand();
	}

	// Sort small arrays with all the algorithms
	fprintf(stdout, "\nRANDOM NUMBERS\n");
	fprintf(stdout, "N insertion_sort selection_sort quick_sort heap_sort merge_sort\n");
	for (size = MAX_DATA_SIZE_Q; size > 0; size -= 2000) {

		fprintf(stdout, "%d", size);

		// Insertion Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		insertionSort(sortedData, size, compareFunc);
		duration = clock() - start;

		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Selection Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		selectionSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Quick Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		quickSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Heap Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		heapSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Merge Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		mergeSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		fprintf(stdout, "\n");
		
	}
	fprintf(stdout, "\n");

	// Sort very large arrays with the n.log(n) algorithms
	fprintf(stdout, "N quick_sort heap_sort merge_sort\n");
	for (size = MAX_DATA_SIZE; size > 0; size -= 50000) {

		fprintf(stdout, "%d", size);

		// Quick Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		quickSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Heap Sort		
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		heapSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Merge Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		mergeSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		fprintf(stdout, "\n");
		
	}

	// Free Memory
	free(initialData);
	free(sortedData);

}

// Run the algorithms on various arrays with already sorted values and time them
void runSortsUsingSortedValues(void)
{
	int i = 0;
	Type value = 0;
	Type *initialData = NULL;
	Type *sortedData = NULL;
	clock_t start = 0;
	clock_t duration = 0;
	int size = 0;
	BOOL (*compareFunc)(Type a, Type b) = intGreaterThan;

	// Memory Allocation for the arrays
	initialData = (Type *)calloc(MAX_DATA_SIZE_Q, sizeof(Type));
	if (initialData == NULL) {
		fprintf(stderr, "Memory Allocation Error!!!\n");
		exit(1);
	}
	sortedData = (Type *)calloc(MAX_DATA_SIZE_Q, sizeof(Type));
	if (sortedData== NULL) {
		fprintf(stderr, "Memory Allocation Error!!!\n");
		exit(1);
	}

	// Initialize the arrays with sorted values
	for (i = 0; i < MAX_DATA_SIZE_Q; i++) {
        initialData[i] = i;
	}

	// Sort the arrays using all the algorithms
	fprintf(stdout, "\nALREADY SORTED NUMBERS\n");
	fprintf(stdout, "N insertion_sort selection_sort quick_sort heap_sort merge_sort\n");
	for (size = MAX_DATA_SIZE_Q; size > 0; size -= 2000) {

		fprintf(stdout, "%d", size);

		// Insertion Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		insertionSort(sortedData, size, compareFunc);
		duration = clock() - start;

		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Selection Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		selectionSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Quick Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		quickSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Heap Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		heapSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Merge Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		mergeSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		fprintf(stdout, "\n");
		
	}

	// Sort small arrays using the n.log(n) algorithms
	fprintf(stdout, "N quick_sort heap_sort merge_sort\n");
	for (size = MAX_DATA_SIZE_Q; size > 0; size -= 2000) {

		fprintf(stdout, "%d", size);

		// Quick Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		quickSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Heap Sort		
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		heapSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		// Merge Sort
		memcpy((void *)sortedData, (const void *)initialData, size * sizeof(Type));
		start = clock();
		mergeSort(sortedData, size, compareFunc);
		duration = clock() - start;
		fprintf(stdout, " %g", (float)duration / (float)CLOCKS_PER_SEC );

		fprintf(stdout, "\n");
		
	}

	// Free memory
	free(initialData);
	free(sortedData);

}